"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

var _fs = require("fs");

var _fs2 = _interopRequireDefault(_fs);

var _path = require("path");

var _path2 = _interopRequireDefault(_path);

function isObject(obj) {
  return obj !== null && typeof obj === "object" && !Array.isArray(obj);
}

function uppercaseFirstChar(str) {
  return str.charAt(0).toUpperCase() + str.slice(1);
}

var generateFragmentFromPropsFunctionName = "generateFragmentFromProps";

exports["default"] = function (_ref) {
  var Plugin = _ref.Plugin;
  var t = _ref.types;

  function flowTypeAnnotationToString(type) {
    if (t.isStringTypeAnnotation(type)) {
      return "string";
    } else if (t.isNumberTypeAnnotation(type)) {
      return "number";
    } else if (t.isBooleanTypeAnnotation(type)) {
      return "boolean";
    }
    return "any";
  }

  // convert an ObjectTypeAnnotation to a js object
  function convertFlowObjectTypeAnnotation(objectType) {
    return objectType.properties.reduce(function (obj, property) {
      var key = property.key.name;
      var value = property.value;
      if (t.isObjectTypeAnnotation(value)) {
        obj[key] = convertFlowObjectTypeAnnotation(value);
      } else {
        obj[key] = flowTypeAnnotationToString(value);
      }
      return obj;
    }, {});
  }

  function convertToGraphQLString(obj) {
    var level = arguments.length <= 1 || arguments[1] === undefined ? 1 : arguments[1];

    var indentation = "  ".repeat(level);

    var strings = Object.keys(obj).reduce(function (parts, key) {
      var value = obj[key];
      var str = key;
      if (isObject(value)) {
        str = str + " {\n" + convertToGraphQLString(value, level + 1) + "\n" + indentation + "}";
      }
      parts.push(indentation + str);
      return parts;
    }, []);

    return strings.join(",\n");
  }

  function storeTypeAlias(node, state) {
    state.opts.extra.flowTypes = state.opts.extra.flowTypes || {};
    state.opts.extra.flowTypes[node.id.name] = node;
  }

  function storeComponentPropType(state, componentName, propType) {
    state.opts.extra.componentPropTypes = state.opts.extra.componentPropTypes || {};
    state.opts.extra.componentPropTypes[componentName] = propType;
  }

  function storeImports(state, filename, variables) {
    state.opts.extra.storedImports = state.opts.extra.storedImports || {};
    variables.forEach(function (variable) {
      return state.opts.extra.storedImports[variable] = filename;
    });
  }

  function storeRelayContainerFragments(state, name, fragments) {
    state.opts.extra.relayContainerFragments = state.opts.extra.relayContainerFragments || {};
    state.opts.extra.relayContainerFragments[name] = fragments;
  }

  function isPluginActive(state) {
    return state.opts.extra.relayFlowQueryPluginActive;
  }

  function extendsReactComponent(node) {
    if (!node || !t.isClassDeclaration(node)) {
      return false;
    }

    var superClass = node.superClass;
    return superClass && superClass.object.name === "React" && superClass.property.name === "Component";
  }

  function isTypeImport(node) {
    return t.isImportDeclaration(node) && (node.importKind === "type" || node.importKind === "typeof");
  }

  function parseImport(node, currentFilename) {
    if (!t.isImportDeclaration(node)) {
      return null;
    }

    if (!currentFilename) {
      throw new Error("Cannot parse import if there is no current file");
    }

    var variables = node.specifiers.map(function (specifier) {
      return specifier.local.name;
    });
    var importFile = node.source.value;
    if (importFile === ".") {
      importFile = "./index";
    }

    var filename = "";
    if (importFile[0] === ".") {
      (function () {
        var importFileName = _path2["default"].parse(importFile).name;
        var dir = _path2["default"].dirname(_path2["default"].resolve(_path2["default"].dirname(currentFilename), importFile));
        var files = _fs2["default"].readdirSync(dir).filter(function (file) {
          var _path$parse = _path2["default"].parse(file);

          var name = _path$parse.name;

          return name === importFileName;
        });
        filename = _path2["default"].resolve(dir, files[0]);
      })();
    } else {
      filename = require.resolve(importFile);
    }

    return { filename: filename, variables: variables };
  }

  function parseReactComponentClass(self, node) {
    if (!t.isClassProperty(node)) {
      return null;
    }

    if (node.key.name !== "props") {
      return null;
    }

    var typeAnnotation = node.typeAnnotation && node.typeAnnotation.typeAnnotation;
    if (!typeAnnotation || !t.isGenericTypeAnnotation(typeAnnotation)) {
      return null;
    }

    var classDec = self.findParent(function (p) {
      return t.isClassDeclaration(p);
    });
    if (!extendsReactComponent(classDec && classDec.node)) {
      return null;
    }

    var className = classDec.node.id.name;
    var propType = typeAnnotation.id.name;

    return { className: className, propType: propType };
  }

  // parse a file with the provided visitor
  function parseFile(filename, visitor, scope) {
    var babel = require("babel-core");
    var program = babel.parse(_fs2["default"].readFileSync(filename, "utf8"));

    babel.traverse(program, visitor, scope, program);
  }

  function isRelayCreateContainer(node) {
    return node.get("callee").matchesPattern("Relay.createContainer");
  }

  return new Plugin("flow-relay-query", {
    visitor: {
      TypeAlias: function TypeAlias(node, parent, scope, state) {
        if (!isPluginActive(state)) {
          return;
        }

        storeTypeAlias(node, state);
      },

      JSXElement: function JSXElement(node, parent, scope, state) {
        if (!isPluginActive(state)) {
          return;
        }

        var name = t.isJSXIdentifier(node.openingElement.name) && node.openingElement.name.name;
        var imports = state.opts.extra.storedImports || {};
        var filename = imports[name];
        if (filename) {
          var _visitor = {
            CallExpression: function CallExpression(n) {
              if (isRelayCreateContainer(this)) {
                var reactComponentName = n.arguments[0].name;
                var fragments = n.arguments[1].properties.filter(function (_) {
                  return _.key.name === "fragments";
                });
                var fragmentNames = fragments.length === 1 ? fragments[0].value.properties.map(function (_) {
                  return _.key.name;
                }) : [];
                storeRelayContainerFragments(state, reactComponentName, fragmentNames);
              }
            }
          };
          parseFile(filename, _visitor, scope);
        }
      },

      ClassProperty: function ClassProperty(node, parent, scope, state) {
        if (!isPluginActive(state)) {
          return;
        }

        var _ref2 = parseReactComponentClass(this, node) || {};

        var className = _ref2.className;
        var propType = _ref2.propType;

        if (className && propType) {
          storeComponentPropType(state, className, propType);
        }
      },

      ImportDeclaration: function ImportDeclaration(node, parent, scope, state) {
        var typeImport = isTypeImport(node);

        // track imports for use later
        if (!typeImport) {
          var _ref3 = parseImport(node, state.opts.filename) || {};

          var filename = _ref3.filename;
          var variables = _ref3.variables;

          storeImports(state, filename, variables);
        }

        // handle imported types
        if (typeImport) {
          (function () {
            var _ref4 = parseImport(node, state.opts.filename) || {};

            var filename = _ref4.filename;
            var variables = _ref4.variables;

            var visitor = {
              TypeAlias: function TypeAlias(n) {
                var typeName = n.id.name;
                if (variables.indexOf(typeName) >= 0) {
                  storeTypeAlias(n, state);
                }
              }
            };
            parseFile(filename, visitor, null);
          })();
        }

        // remove the marker function import
        var remove = node.specifiers.filter(function (specifier) {
          return specifier.local.name === generateFragmentFromPropsFunctionName;
        }).length > 0;

        if (remove) {
          state.opts.extra.relayFlowQueryPluginActive = true;
          this.dangerouslyRemove();
        }
      },

      ArrowFunctionExpression: function ArrowFunctionExpression(node, parent, scope, state) {
        if (!isPluginActive(state)) {
          return;
        }

        var functionName = parent.id && parent.id.name;

        if (functionName && node.params.length === 1) {
          var typeAnnotation = node.params[0].typeAnnotation;
          typeAnnotation = typeAnnotation && typeAnnotation.typeAnnotation;

          if (typeAnnotation && t.isGenericTypeAnnotation(typeAnnotation)) {
            var propType = typeAnnotation.id.name;
            storeComponentPropType(state, functionName, propType);
          }
        }
      },

      CallExpression: function CallExpression(node, parent, scope, state) {
        if (!isPluginActive(state)) {
          return;
        }

        if (!t.isIdentifier(node.callee, { name: generateFragmentFromPropsFunctionName })) {
          return;
        }

        var relayCreateContainer = this.findParent(function (p) {
          return t.isCallExpression(p) && isRelayCreateContainer(p);
        });
        if (!relayCreateContainer) {
          return;
        }

        var reactComponentName = relayCreateContainer.node.arguments[0].name;
        var typeName = state.opts.extra.componentPropTypes && state.opts.extra.componentPropTypes[reactComponentName];
        if (!typeName) {
          throw new Error("React component " + reactComponentName + " does not have flow typed props");
        }

        var type = state.opts.extra.flowTypes && state.opts.extra.flowTypes[typeName];
        if (!type) {
          throw new Error("There is no flow type with name " + typeName);
        }

        var fragmentKey = t.isProperty(parent) && parent.key.name;
        var typeProperties = type.right.properties;

        var typeAtKey = null;
        for (var i = 0; i < typeProperties.length; i++) {
          var typeProperty = typeProperties[i];
          if (t.isObjectTypeProperty(typeProperty) && typeProperty.key.name === fragmentKey) {
            typeAtKey = typeProperty.value;
          }
        }

        if (!typeAtKey) {
          throw new Error("There is no property named '" + fragmentKey + "' in flow type " + typeName);
        }

        var obj = t.isObjectTypeAnnotation(typeAtKey) ? convertFlowObjectTypeAnnotation(typeAtKey) : {};
        var graphQlQueryBody = convertToGraphQLString(obj);

        var childRelayContainers = state.opts.extra.relayContainerFragments || {};
        var childRelayContainersForFragment = Object.keys(childRelayContainers).reduce(function (c, name) {
          var fragmentKeys = childRelayContainers[name];
          if (fragmentKeys.indexOf(fragmentKey) >= 0) {
            c.push(name);
          }
          return c;
        }, []);

        var fragmentName = node.arguments[0] && node.arguments[0].value || uppercaseFirstChar(fragmentKey);
        var graphQlQuery = "\n" + "fragment on " + fragmentName + " {\n" + graphQlQueryBody;
        var graphQlQueryEnd = "\n}\n";

        if (childRelayContainersForFragment.length > 0) {
          if (graphQlQueryBody) {
            graphQlQuery = graphQlQuery + ",\n  ";
          }

          graphQlQuery = graphQlQuery + childRelayContainersForFragment.map(function (name) {
            return "${" + name + ".getFragment('" + fragmentKey + "')}";
          }).join(",\n  ");
        }
        graphQlQuery = graphQlQuery + graphQlQueryEnd;

        this.replaceWithSourceString("() => Relay.QL`" + graphQlQuery + "`");
      }
    }
  });
};

module.exports = exports["default"];